/*
 * Copyright 2012-2013 Brian L. Browning
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package vcf;

import blbutil.SampleFileIterator;
import beagleutil.ChromInterval;
import beagleutil.Samples;
import blbutil.Filter;
import ibd.IntPairToIntMap;
import java.io.File;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import main.ExcludeIdFilter;
import main.Marker;

/* Code review on 03 Oct 2013 */

/**
 * Class <code>VcfMarkerData</code> represents genotype data (GT FORMAT field)
 * in a VCF file together with per-marker missing and allele count summary
 * statistics.
 *
 * @author Brian L. Browning <browning@uw.edu>
 */
public final class VcfMarkerData {

    private static final int INIT_N_MARKERS = 100000;
    private static final float LOAD_FACTOR = 0.75f;
    public static final int NULL_VALUE = -79;

    private final int nInitialMarkers; // includes markers failing MAF filter
    private final List<Marker> freqExcludedMarkers;
    private final List<MarkerData> dataList;
    private final BitSet isCorrelated;
    private final Samples samples;

    /**
     * Constructs a <code>VcfMarkerData</code> instance for the specified
     * genotype data.
     *
     * @param gtFile VCF file with GT FORMAT fields for each VCF record.
     * @param usePhase <code>true</code> if phase information in the specified
     * VCF file record will be used, and <code>false</code> if phase
     * information in the specified VCF file record will be ignored.
     * @param sampleExclusions samples to be excluded.
     * @param markerExclusions markers to be excluded.
     * @param chromInterval a string representing a marker filter.
     * @param minAlleleCount minimum count required for at least two of
     * a marker's alleles in order for the marker to pass the
     * minor allele frequency filter.
     * @param r2Window number of markers in sliding marker window used
     * to detect high-correlated alleles  Pairs of markers with indices whose
     * difference exceeds <code>r2Window</code> are assumed to
     * have correlation less than or equal to the specified
     * <code>r2Max</code> value.  Marker indices are assigned after excluding
     * markers in the specified excluded markers file and after excluding
     * markers failing the minor allele frequency filter (see
     * <code>minAlleleCount</code> parameter).
     * @param r2Max maximum squared correlation permitted between
     * any two markers.
     *
     * @throws NullPointerException if <code>gtFile==null/code>.
     *
     * @throws IllegalArgumentException if the specified <code>gtFile</code>
     * does not conform to the specifications for a VCF file, if
     * <code>MarkerFilter.isValidRegion(markerFilterString)==false</code>,
     * or if <code>minAlleleCount < 0 || r2Window < 0 || r2Max < 0.0f
     * || r2Max > 1.0f || float.isNaN(r2Max)</code>.
     */
    public VcfMarkerData(File gtFile, boolean usePhase, File exclMarkersFile,
            File exclSamplesFile, String chromInterval,
            int minAlleleCount, int r2Window, float r2Max) {
        if (minAlleleCount < 0) {
            throw new IllegalArgumentException("minAlleles: " + minAlleleCount);
        }
        if (r2Window < 0) {
            throw new IllegalArgumentException("r2Window: " + r2Window);
        }
        if (r2Max < 0.0f || r2Max > 1.0f || Float.isNaN(r2Max)) {
            throw new IllegalArgumentException("r2Max: " + r2Max);
        }
        SampleFileIterator<VcfRecord> it = vcfIterator(gtFile, exclMarkersFile,
                exclSamplesFile, chromInterval);
        this.samples = it.samples();
        this.freqExcludedMarkers = new ArrayList<Marker>(1000);
        this.dataList = new ArrayList<MarkerData>(INIT_N_MARKERS);
        this.isCorrelated = new BitSet(INIT_N_MARKERS);

        this.nInitialMarkers = readData(it, usePhase, minAlleleCount, r2Window,
                r2Max, dataList, isCorrelated, freqExcludedMarkers);
        it.close();
    }

    private static SampleFileIterator<VcfRecord> vcfIterator(File gtFile,
            File exclMarkersFile, File exclSamplesFile, String chromInterval) {
        Filter<String> sampleFilter = VcfUtils.ACCEPT_ALL_FILTER;
        if (exclSamplesFile!=null) {
            sampleFilter = new VcfUtils.ExcludeSampleFilter(exclSamplesFile);
        }
        ChromInterval ci = ChromInterval.parse(chromInterval);
        SampleFileIterator<VcfRecord> it = new VcfIterator(gtFile, sampleFilter);
        if (ci != null) {
            it = new IntervalVcfIterator(it, ci);
        }
        if (exclMarkersFile!=null) {
            Filter<Marker> markerFilter = new ExcludeIdFilter(exclMarkersFile);
            it = new FilteredVcfIterator(it, markerFilter);
        }
        return it;
    }

    private static int readData(SampleFileIterator<VcfRecord> it, boolean usePhase,
            int minAlleleCount, int r2Window, float r2Max,
            List<MarkerData> dataList, BitSet isExcluded,
            List<Marker> lowAlleleList) {
        int markerCnt = 0;
        int chromIndex = -1;
        while (it.hasNext()) {
            ++markerCnt;
            VcfMarker obsData = new LowMemGT(it.next(), usePhase);
            if (chromIndex == -1) {
                chromIndex = obsData.marker().chromIndex();
            }
            if (obsData.marker().chromIndex()!=chromIndex) {
                break;
            }
            MarkerData mData = new MarkerData(obsData);
            byte minor = mData.minorAllele();
            if (minor != -1 && mData.alleleCount(minor) >= minAlleleCount) {
                dataList.add(mData);
                if (r2Window>0) {
                    checkLastMarkerR2(r2Window, r2Max, dataList, isExcluded);
                }
            }
            else {
                lowAlleleList.add(mData.marker());
            }
        }
        return markerCnt;
    }

    private static void checkLastMarkerR2(int r2Window, float r2Max,
            List<MarkerData> dataList, BitSet isExcluded) {
        int radius = 1 + r2Window/2;
        int startIndex = Math.max(0, dataList.size() - radius);
        MarkerData mdB = dataList.get(dataList.size()-1);
        byte alleleB = mdB.majorAllele();
        boolean finished = false;
        for (int k=dataList.size()-2; k>=startIndex && finished==false; --k) {
            if (isExcluded.get(k) == false) {
                MarkerData mdA = dataList.get(k);
                byte alleleA = mdA.majorAllele();
                double r2 = MarkerData.r2(mdA, mdB, alleleA, alleleB);
                if (r2 > r2Max) {
                    if (mdB.alleleFrequency(alleleB) > mdA.alleleFrequency(alleleA)) {
                        isExcluded.set(k);
                    }
                    else {
                        isExcluded.set(dataList.size()-1);
                        finished = true;
                    }
                }
            }
        }
    }

    /**
     * Returns the number of markers read from the VCF file (including
     * markers failing the minimum allele carrier filter).
     */
    public int nInitialMarkers() {
        return nInitialMarkers;
    }

    /**
     * Returns the number of markers excluded by the minor allele frequency
     * filter.
     * @return the number of markers excluded by the minor allele frequency
     * filter.
     */
    public int nFrequencyExcludedMarkers() {
        return freqExcludedMarkers.size();
    }

    /**
     * Returns the specified marker excluded by the minor allele frequency
     * filter.
     * @param index the index in the list of markers that failed
     * the minor allele frequency filter.
     * @return the specified marker excluded by the minor allele frequency
     * filter.
     * @throws IndexOutOfBoundsException if
     * <code>index < 0 || index &ge; this.nFrequencyExcludedMarkers</code>.
     */
    public Marker frequencyExcludedMarkers(int index) {
        return freqExcludedMarkers.get(index);
    }

    /**
     * Returned the number of markers.  The returned count does not include
     * markers failing the minor allele frequency filter.
     * @return the number of markers.
     */
    public int nMarkers() {
        return dataList.size();
    }

    /**
     * Returns the number of samples represented by <code>this</code>.
     * @return the number of samples represented by <code>this</code>.
     */
    public int nSamples() {
        return samples.nSamples();
    }

    /**
     * Returns the samples represented by <code>this</code>.
     * @return the samples represented by <code>this</code>.
     */
    public Samples samples() {
        return samples;
    }

    /**
     * Returns the specified marker data.
     * @param index a marker index.
     * @return the specified marker data.
     * @throws IndexOutOfBoundsException if
     * <code>index < 0 || index &ge; this.nMarkers()</code>.
     */
    public MarkerData get(int index) {
        return dataList.get(index);
    }

    /**
     * Returns <code>true</code> if the specified marker must be excluded
     * to remove remove excess inter-marker correlation, and returns
     * <code>false</code> otherwise.
     *
     * @param index a marker index.
     *
     * @return <code>true</code> if the specified marker must be excluded
     * to remove remove excess inter-marker correlation, and returns
     * <code>false</code> otherwise.
     *
     * @throws IndexOutOfBoundsException if
     * <code>index < 0 || index &ge; this.nMarkers()</code>.
     */
    public boolean isCorrelated(int index) {
        return isCorrelated.get(index);
    }

    /**
     * Returns the number of markers that must be excluded to remove
     * excess inter-marker correlation.
     * @return the number of markers that must be excluded to remove
     * excess inter-marker correlation.
     */
    public int nCorrelatedMarkers() {
        return isCorrelated.cardinality();
    }

    /**
     * Returns a map of genome coordinates to marker indices.  The
     * genome coordinates are an ordered pair: (chromosome index, position).
     * The marker indices are determined after excluding markers
     * failing the minor allele frequency filter.
     *
     * @return a map of genome coordinates to marker indices.
     */
    public IntPairToIntMap markerIndexMap() {
        IntPairToIntMap map = new IntPairToIntMap(INIT_N_MARKERS,
                LOAD_FACTOR, NULL_VALUE);
        for (int j=0, n=dataList.size(); j<n; ++j) {
            Marker m = dataList.get(j).marker();
            map.put(m.chromIndex(), m.pos(), j);
        }
        return map;
    }
}
